#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/sched.h>

#ifdef K2_2
#include <asm/uaccess.h>
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <stdio.h>
#include "dskc31_ioctl.h"

/* #define BASE 0x378 */
#define CONTROL BASE+2
#define STATUS 	BASE+1

#ifdef PPI
#define PA PPIBASE
#define PB PPIBASE+1
#define PC PPIBASE+2
#define CNTL PPIBASE+3
#define MPAOUT 0x82
#define MPAIN  0x92
#define ALLIN  0x9B

#define DSK_RESET 7			/* PC pin for /RESET */
#define DSK_STB   6			/* PC pin for /HPSTB */
#define DSK_ACK   0			/* PB pin for /HPACK */

#define RESL   (DSK_RESET*2 + 1)	/* /RESET LOW  */
#define RESH   (DSK_RESET*2)		/* /RESET HIGH */
#define STBL   (DSK_STB*2)		/* /HPSTB LOW  */
#define STBH   (DSK_STB*2 + 1)		/* /HPSTB HIGH */
#define ACKM   (1<<DSK_ACK)		/* ACK MASK    */
#define RNRES  
#endif

/*#define MN 99			/* MAJOR NUMBER */

#define C31_TIMEOUT	50
#define C31_OFF		20
#define STBDEL		4	/* Strobe edge delay */

static int c31_major;
static int n_open;

/************ funciones auxiliares *************/

void mydelay(int n) {
#ifdef	K2_2
	schedule_timeout(n);
#else
	current->state=TASK_INTERRUPTIBLE;
	current->timeout=jiffies+n;
	schedule();
#endif
}

int dsp_reset()
{
#ifdef PPI
	outb(MPAOUT,CNTL);
	outb(RESL,CNTL);	    /* RESET Low & STB High */
	outb(STBH,CNTL);
	mydelay(5);
	if (!(inb(PB)&ACKM)) { /* -ACK mut be high */
		printk("DSK_Reset: DSK not ready\n");
		return -1;
	}
	outb(RESH,CNTL);	/* RESET High */
	mydelay(5);
	outb(STBL,CNTL);	/* -STRB L: bootloader init */
	mydelay(5);
	if (inb(PB)&ACKM) { /* -ACK mut be low */
		printk("DSK_Reset: DSK not ready\n");
		return -1;
	}
	outb(RESL,CNTL); /* reset it again */
	outb(STBH,CNTL);
	mydelay(5);
	outb(RESH,CNTL);
	return 0;

#else
	outb(0,CONTROL);	    /* RESET active */
	mydelay(5);
	if (!(inb(STATUS)&8)) { /* -ACK mut be high */
		printk("DSK_Reset: DSK not ready\n");
		return -1;
	}
	outb(0x4,CONTROL);	/* -STRB H */
	mydelay(5);
	outb(5,CONTROL);	/* -STRB L: bootloader init */
	mydelay(5);
	if (inb(STATUS)&8) { /* -ACK mut be low */
		printk("DSK_RESET: DSK not ready\n");
		return -1;
	}
	outb(0x0,CONTROL); /* reset it again */
	mydelay(5);
	outb(0x04,CONTROL);
#endif
	return 0;
}


/*********** Kernel Interface **************/
#ifdef K2_2
static ssize_t read_c31(struct file * file, char * buf,
                        size_t count, loff_t *ppos)
#else
static int read_c31(struct inode *inode, struct file *file, char *buf, int count)
#endif
{
	int i,left,tim,nd;
	unsigned char a,b;
	
	if (verify_area(VERIFY_WRITE,buf,count)==-EFAULT) return -EFAULT;

#ifdef	PPI
	outb(MPAIN,CNTL);	/* Port A IN */

	for (left=count;left>0;left--) {
	
		tim=C31_TIMEOUT; nd=C31_OFF;
		while(((inb(PB))&ACKM)!=0) {	/* wait until ACK */
			tim --; 
			if (!tim) { 
				tim=C31_TIMEOUT;
				mydelay(1);
				nd--;
				if (!nd) return(0);
			}
		} 
		a=inb(PA);
		for (i=0;i<STBDEL;i++) outb(STBL,CNTL); /* Strobe LOW */
		for (i=0;i<STBDEL;i++) outb(STBH,CNTL); /* Strobe HIGH: -READY pulse */

		put_user(a,buf++);
	}

#else
	for (left=count;left>0;left--) {
		tim=C31_TIMEOUT; nd=C31_OFF;
		while(((inb(STATUS))&8)!=0) {	/* Wait until ACK */
			tim --; 
			if (!tim) { 
				tim=C31_TIMEOUT;
				mydelay(1);
				nd--;
				if (!nd) return(0);
			}
		}

		for (i=0;i<STBDEL;i++) outb(0x5,CONTROL); /* Strobe LOW */
		a=inb(STATUS);
		for (i=0;i<STBDEL;i++) outb(0x4,CONTROL); /* Strobe HIGH: -READY pulse */

		a^=0x80; a>>=4;

		tim=C31_TIMEOUT; nd=C31_OFF;
		while(((inb(STATUS))&8)!=0) {	/* wait until ACK */
			tim --; 
			if (!tim) { 
				tim=C31_TIMEOUT;
				mydelay(1);
				nd--;
				if (!nd) return(0);
			}
		}

		for (i=0;i<STBDEL;i++) outb(0x5,CONTROL); /* Strobe LOW */
		b=inb(STATUS);
		for (i=0;i<STBDEL;i++) outb(0x4,CONTROL); /* Strobe HIGH: -READY pulse */

		b^=0x80; b&=0xf0;
		a|=b;
		put_user(a,buf++);
	}
#endif
	return count;
}


#ifdef K2_2
static ssize_t write_c31(struct file * file, const char * buf,
                        size_t count, loff_t *ppos)
#else
static int write_c31(struct inode *node, struct file *file, const char *buf, int count)
#endif
{
	int i,left,tim,nd;
	unsigned char a;

	if (verify_area(VERIFY_READ,buf,count)==-EFAULT) return -EFAULT;

#ifdef PPI
	outb(MPAOUT,CNTL);	/* Port PA OUT */
	
	for (left=count;left>0;left--) {
#ifdef K2_2
		if(get_user(a,buf++)) return -EFAULT;
#else
		a=get_user(buf++);
#endif
		outb(a,PA);
		for (i=0;i<STBDEL;i++) outb(STBL,CNTL); /* Strobe LOW */
		tim=C31_TIMEOUT; nd=C31_OFF;
		while((inb(PB)&ACKM)!=0) {
			tim --; 
			if (!tim) { 
				tim=C31_TIMEOUT;
				mydelay(1);
				nd--;
				if (!nd) return(0);
			}

		}
		for (i=0;i<STBDEL;i++) outb(STBH,CNTL); /* Strobe HIGH */
	}

#else
	outb(0x4,CONTROL);		/* PS2 port OUT */
	
	for (left=count;left>0;left--) {
#ifdef K2_2
		if(get_user(a,buf++)) return -EFAULT;
#else
		a=get_user(buf++);
#endif
		outb(a,BASE);
		for (i=0;i<STBDEL;i++) outb(0x5,CONTROL); /* Strobe LOW */
		tim=C31_TIMEOUT; nd=C31_OFF;
		while((inb(STATUS)&8)!=0) {
			tim --; 
			if (!tim) { 
				tim=C31_TIMEOUT;
				mydelay(1);
				nd--;
				if (!nd) return(0);
			}
		}
		for (i=0;i<STBDEL;i++) outb(0x4,CONTROL); /* Strobe HIGH */
	}
#endif
	return count;
}

#ifdef K2_2
static long long lseek_c31(struct file * file, long long offset, int origin)
#else
static int lseek_c31(struct inode *node, struct file *file, off_t offset, int orig)
#endif
{
	return file->f_pos=0;
}

static int open_c31(struct inode *node, struct file *file)
{

	if (n_open) return -EBUSY;

	MOD_INC_USE_COUNT;
	n_open++;
	
	return 0;
}

#ifdef K2_2
static int release_c31(struct inode *node, struct file *file)
#else
static void release_c31(struct inode *node, struct file *file)
#endif
{
	MOD_DEC_USE_COUNT;
	n_open--;
#ifdef	PPI
	outb(ALLIN,CNTL);	/* All pins IN */
#else
	outb(0,BASE);	/* DSK can be powered from port otherwise */
#endif
#ifdef	K2_2
	return 0;
#endif
}

static int ioctl_c31(struct inode *node, struct file *file, unsigned int code, unsigned long pointer)
{
	int ret;
	
	switch (code) {
	case 0:			break;
	case DSK_IOCTL_RESET:	ret=dsp_reset();
				break;
	}
	
	return ret;
}


static struct file_operations c31_fops={
	lseek_c31,
	read_c31,
	write_c31,
	NULL,
	NULL,
	ioctl_c31,
	NULL,
	open_c31,
#ifdef	K2_2
	NULL,
#endif
	release_c31
};

#ifdef PPI
int probe_PPI()
{
	outb(MPAIN,CNTL);	/* All PC pins OUT */
	if (inb(PC)) return 1;
	outb(11,CNTL);		/* Bit Set PC5 */
	if (inb(PC)!=0x20) return 1;
	outb(ALLIN,CNTL);	/* All pins IN */
	return 0;
}
#endif

int init_module(void)
{
	printk("Loading TMS320C31 DSK module ...\n");
	if ((c31_major=register_chrdev(MN,"dskc31",&c31_fops))==-EBUSY) {
		printk("Unable to get major for c31 device\n");
		return -EIO;
	}

#ifdef	PPI
	if (check_region(PPIBASE,4)==-EBUSY) {
		 printk("I/O range in use (%0x)\n",PPIBASE);
		 return -EIO;
	}

	request_region(PPIBASE,4,"dskc31");
	if (!probe_PPI()) printk("DSKC31: PPI Interface at 0x%X\n",PPIBASE);
	else printk("DSKC31: PPI probe failed\n");
#else
	if (check_region(BASE,8)==-EBUSY) {
		 printk("I/O range in use (%0x)\n",BASE);
		 return -EIO;
	}

	request_region(BASE,8,"dskc31");
	printk("DSKC31: compiled for LPT Interface at 0x%X\n",BASE);
#endif
	n_open=0;

	return 0;
}

void cleanup_module(void)
{
	unregister_chrdev(MN,"dskc31");
#ifdef	PPI
	release_region(PPIBASE,4);
#else
	release_region(BASE,8);
#endif

}

