#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <stdio.h>
#include "dskc31_ioctl.h"

/* #define BASE 0x1b0 */
#define PA BASE
#define PB BASE+1
#define PC BASE+2
#define CNTL BASE+3

/**********************************************************************
			PROGRAMACION PPI 8255
			
		BIT
		
		0 - PC (lower)	0=out 1=in
		1 - PB		0=out 1=in
		2 - PB mode	(0,1)
		3 - PC (upper)	0=out 1=in
		4 - PA		0=out 1=in
		5,6 - PA mode	(0,1,2)
		7 - mode set	0=write bit 1=mode set

***********************************************************************/		

#define MPAOUT 0x82
#define MPAIN  0x92

/*#define MN 99			/* MAJOR NUMBER */

#define C31_TIMEOUT	50;


static int c31_major;
static int n_open;

/************ funciones auxiliares *************/

void mydelay(int n) {
	current->state=TASK_INTERRUPTIBLE;
	current->timeout=jiffies+n;
	schedule();
}

int dsp_reset()
{
	/* outb(MPAOUT,CNTL); */
	outb(0xC0,PC);	    /* RESET activo */
	mydelay(5);
	if (!(inb(PB)&1)) { /* -ACK mut be high */
		printk("Error: DSK not ready\n");
		return -1;
	}
	outb(0x40,PC);	/* -STRB H */
	mydelay(5);
	outb(0,PC);	/* -STRB L: bootloader init */
	mydelay(5);
	if (inb(PB)&1) { /* -ACK mut be low */
		printk("Error: DSK not ready\n");
		return -1;
	}
	outb(0xC0,PC); /* reset it again */
	mydelay(5);
	outb(0x40,PC);
	return 0;
}


/*********** Interface con el Kernel **************/

static int read_c31(struct inode *inode, struct file *file, char *buf, int count)
{
	int left,tim;
	char a;
	
	if (verify_area(VERIFY_WRITE,buf,count)==-EFAULT) return -EFAULT;
	
	outb(MPAIN,CNTL);	/* Puerto PA como entrada */

	for (left=count;left>0;left--) {
	
		tim=C31_TIMEOUT;
		while(((inb(PB))&1)!=0) {	/* Espera ACK activo */
			tim --; if (tim==0) { tim=C31_TIMEOUT; mydelay(1); }
		} 
		a=inb(PA);
		outb(0,PC); /* Strobe activo */
		outb(0x40,PC); /* Strobe inactivo: pulso -READY*/

		put_user(a,buf++);
	}
	return count;
}

static int write_c31(struct inode *node, struct file *file, char *buf, int count)
{
	int left,tim;
	char a;
	
	if (verify_area(VERIFY_READ,buf,count)==-EFAULT) return -EFAULT;

	outb(MPAOUT,CNTL);	/* Puerto PA como salida */
	
	for (left=count;left>0;left--) {
		a=get_user(buf++);

		outb(a,PA);
		outb(0,PC); /* Strobe Activo */
		tim=C31_TIMEOUT;
		while((inb(PB)&1)!=0) {
			tim--; if (tim==0) { tim=C31_TIMEOUT; mydelay(1); }
		}
		outb(0x40,PC); /* Strobe inactivo */
	}
	return count;
}

static int lseek_c31(struct inode *node, struct file *file, off_t offset, int orig)
{
	return file->f_pos=0;
}

static int open_c31(struct inode *node, struct file *file)
{

	if (n_open) return -EBUSY;

	MOD_INC_USE_COUNT;
	n_open++;
	
	return 0;
}

static void release_c31(struct inode *node, struct file *file)
{
	MOD_DEC_USE_COUNT;
	n_open--;
}

static int ioctl_c31(struct inode *node, struct file *file, unsigned int code, unsigned long pointer)
{
	int ret;
	switch (code) {
	case 0:			break;
	case DSK_IOCTL_RESET:	ret=dsp_reset();
				break;
	}
	return ret;
}


static struct file_operations c31_fops={
	lseek_c31,
	read_c31,
	write_c31,
	NULL,
	NULL,
	ioctl_c31,
	NULL,
	open_c31,
	release_c31,
};

#ifndef MODULE
long c31_init(long mem_start, long mem_end)
{
	if (c31_major=register_chrdev(MN,"c31",&c31_fops))
		printk("Unable to get major for c31 device\n");
	return mem_start;
}
#else
int init_module(void)
{
	if ((c31_major=register_chrdev(MN,"c31",&c31_fops))==-EBUSY) {
		printk("Unable to get major for c31 device\n");
		return -EIO;
	}
	n_open=0;

	if (check_region(BASE,4)==-EBUSY) {
		 printk("I/O range in use\n");
		 return -EIO;
	}
	request_region(BASE,4,"dskc31");

	outb(MPAIN,CNTL);	/* Puerto PA como entrada */

	return 0;
}

void cleanup_module(void)
{
	unregister_chrdev(MN,"c31");
	release_region(BASE,4);
}

# endif
